###############################################################################
# Coscine Python SDK
# Copyright (c) 2020-2023 RWTH Aachen University
# Licensed under the terms of the MIT License
# For more information on Coscine visit https://www.coscine.de/.
###############################################################################


"""
Stateful integration test, KISS.
"""


from datetime import datetime
import logging
import traceback
from os import getenv
from concurrent.futures import ThreadPoolExecutor, Future, wait
import coscine

###############################################################################
# Global variables
###############################################################################


logger = logging.getLogger(__name__)
exceptions: list[str] = []
profiles: list[str] = []
good = 0
bad = 0
errid = 1

MARKUP: str = """
# Coscine Python SDK Test Report
This report was automatically generated by the Coscine Python SDK
integration test.

## Application Profiles
Showcases whether the Coscine Python SDK is able to interface with
the listed application profile. All application profiles marked with
✔ have been tested and considered working, while all profiles marked
with ❌ may bring along issues. That does not mean that those will
not work with the Python SDK, just that functionality may be limited
or that the application profile itself is broken. The integration tests
are very strict on whether a profile is considered working.

### Statistics

| Property | Value   |
|----------|---------|
| Total    | {total} |
| ✔        | {good}  |
| ❌       | {bad}   |

### Test Results

|Name | Link | Supported| EID |
|-----|------|----------|-----|
{profiles}

## Exceptions
Lists the exceptions that were raised during testing.  

{exceptions}

"""


def report_result(
    profile: coscine.ApplicationProfileInfo,
    works: bool,
    err: Exception | None = None
) -> None:
    """
    """
    global good
    global bad
    global errid
    if works:
        good += 1
        eid = " "
    else:
        bad += 1
        message = "".join(traceback.format_exception(err))
        logger.error(message)
        eid = f"{errid}"
        errid += 1
        exceptions.append(
            f"### Exception EID:{eid}:  \n"
            f"{message}"
        )
    profiles.append(
        f"{profile.name} | "
        f"[{profile.uri}]({profile.uri}) | "
        f"{'✔' if works else '❌'} | "
        f"{eid} |  "
    )

###############################################################################
# Setup
###############################################################################


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("test")
logger.info("Initializing Api Client...")
token = getenv("COSCINE_API_TOKEN")
if not token:
    raise ValueError(
        "No Coscine API Token specified! "
        "Create a new environment variable with "
        "the name COSCINE_API_TOKEN!"
    )
client = coscine.ApiClient(token, verbose=False)
logger.info("Creating root test project...")
try:
    project = client.project("coscine-python-sdk-test")
    project.delete()
except coscine.CoscineException:
    pass
project = client.create_project(
    "coscine-python-sdk-test",
    "coscine-python-sdk-test",
    "This project was automatically generated by the "
    "Coscine Python SDKs test script. If you are still able "
    "to see this project you may go ahead and delete it.",
    datetime.now().date(),
    datetime.now().date(),
    "Test User",
    [client.discipline("Computer Science 409")],
    [client.organization("https://ror.org/04xfq0f34")],
    # -> RWTH Aachen University
    client.visibility("Project Members")
)

###############################################################################
#
###############################################################################

for language in client.languages():
    print(language.abbreviation)
    print(language.id)
    print(language.name)
for visibility in client.visibilities():
    print(visibility.id)
    print(visibility.name)
    print(visibility.serialize())
for discipline in client.disciplines():
    print(discipline.id)
    print(discipline.name)
    print(discipline.uri)
for apitoken in client.api_tokens():
    print(apitoken.created)
    print(apitoken.expired)
    print(apitoken.expires)
    print(apitoken.id)
    print(apitoken.name)
    print(apitoken.owner)
for license in client.licenses():
    print(license.id)
    print(license.name)
    print(license.serialize())
for role in client.roles():
    print(role.description)
    print(role.id)
    print(role.name)
for rtype in client.resource_types():
    print(rtype.active)
    print(rtype.general_type)
    print(rtype.specific_type)
    print(rtype.id)

# how to get user via user id?
assert client.validate_pid("http://hdl.handle.net/21.11102/77c4a372-9dce-441c-bf41-7b731a2120f7") is True
assert client.validate_pid("21.11102/77c4a372-9dce-441c-bf41-7b731a2120f7") is True
assert client.validate_pid("garbage") is False

###############################################################################
# Member Management
###############################################################################

logger.info("Testing member management...")
for user in client.users("coscine@protonmail.com")[:3]:
    if user.id != client.self().id:
        project.add_member(user, client.role("Member"))
    logger.info(f"-> Added {user.display_name} to {project.name}!")
for member in project.members():
    print(member.id)
    print(member.project)
    print(member.role)
    print(member.user)
    if member.user.id != client.self().id:
        project.remove_member(member)
project.invite("coscine@protonmail.com", client.role("Guest"))
project.invitations()

###############################################################################
# Resources and Application Profiles
###############################################################################


def test_application_profile(
    client: coscine.ApiClient,
    project: coscine.Project,
    profile: coscine.ApplicationProfileInfo,
    index: int,
    length: int
) -> None:
    """
    Tests an application profile.
    """
    logger.info(f"Testing profile [{index}/{length}] -> {profile.name}")
    resource = project.create_resource(
        profile.name,
        profile.name[:24],
        profile.description or "No description set.",
        client.license("MIT License"),
        client.visibility("Project Members"),
        [client.discipline("Computer Science 409")],
        client.resource_type("linked"),
        0,
        profile
    )
    print(resource.project)
    print(resource.name)
    print(resource.display_name)
    print(resource.id)
    print(resource.access_url)
    print(resource.application_profile)
    print(resource.creator)
    print(resource.metadata())
    print(resource.metadata_form())
    print(resource.created)
    print(resource.description)
    print(resource.disciplines)
    print(resource.files())
    print(resource.graph())
    print(resource.license)
    print(resource.keywords)
    print(resource.archived)
    print(resource.pid)
    print(resource.options)
    print(resource.type)
    print(resource.quota)
    print(resource.usage_rights)
    is_working: bool = True
    for _ in range(3):
        try:
            form = resource.metadata_form()
            form.test()
            str(form)
            form.serialize("testfile.txt")
            form.validate()
            resource.upload("testfile.txt", "http://example.org", form)
            report_result(profile, is_working)
            break
        except ConnectionError:
            logger.info(f"Timeout on {profile.name} - retrying...")
        except Exception as err:
            is_working = False
            report_result(profile, is_working, err)
            break
    resource.delete()


logger.info("Testing all application profiles...")
application_profiles = client.application_profiles()
length = len(application_profiles)
futures: list[Future] = []
with ThreadPoolExecutor(10) as executor:
    for index, profile in enumerate(application_profiles):
        futures.append(executor.submit(
            test_application_profile,
            client,
            project,
            profile,
            index,
            length
        ))
wait(futures)
for future in futures:
    future.result()

###############################################################################
# Generate Test report
###############################################################################

data = MARKUP.format(
    total=len(profiles),
    good=good,
    bad=bad,
    profiles="\n".join(profiles),
    exceptions="\n".join(exceptions)
)
with open("report.md", "w", encoding="utf-8") as fp:
    fp.write(data)

###############################################################################
# Clean Up
###############################################################################

project.delete()
